local function P_IsClimbingValid(player, angle)

	local platx, platy
	local glidesector
	local floorz, ceilingz
	local mo = player.mo
	local rover

	platx = P_ReturnThrustX(mo, angle, mo.radius + FixedMul(8*FRACUNIT, mo.scale))
	platy = P_ReturnThrustY(mo, angle, mo.radius + FixedMul(8*FRACUNIT, mo.scale))

	glidesector = R_PointInSubsector(mo.x + platx, mo.y + platy).sector

	floorz = glidesector.floorheight
	ceilingz = glidesector.ceilingheight

	if (glidesector ~= mo.subsector.sector)
	
		local floorclimb = false
		local topheight, bottomheight

		for rover in glidesector.ffloors()
		
			if (not (rover.flags & FF_EXISTS) or not (rover.flags & FF_BLOCKPLAYER))
				continue
			end

			topheight = rover.topheight
			bottomheight = rover.bottomheight

			floorclimb = true

			if (mo.eflags & MFE_VERTICALFLIP)
			
				if ((topheight < mo.z + mo.height) and ((mo.z + mo.height + mo.momz) < topheight))
					floorclimb = true
				end
				if (topheight < mo.z)
					floorclimb = false
				end
				if (bottomheight > mo.z + mo.height - FixedMul(16*FRACUNIT,mo.scale))
					floorclimb = false
				end
			else
			
				if ((bottomheight > mo.z) and ((mo.z - mo.momz) > bottomheight))
					floorclimb = true
				end
				if (bottomheight > mo.z + mo.height)
					floorclimb = false
				end
				if (topheight < mo.z + FixedMul(16*FRACUNIT,mo.scale))
					floorclimb = false
				end
			end

			if (floorclimb)
				break
			end
		end

		if (mo.eflags & MFE_VERTICALFLIP)
		
			if ((floorz <= mo.z + mo.height)
				and ((mo.z + mo.height - mo.momz) <= floorz))
				floorclimb = true
			end

			if ((floorz > mo.z)
				and glidesector.floorpic == "F_SKY1")
				return false
			end

			if ((mo.z + mo.height - FixedMul(16*FRACUNIT,mo.scale) > ceilingz)
				or (mo.z + mo.height <= floorz))
				floorclimb = true
			end
		else
		
			if ((ceilingz >= mo.z)
				and ((mo.z - mo.momz) >= ceilingz))
				floorclimb = true
			end

			if ((ceilingz < mo.z+mo.height)
				and glidesector.ceilingpic == "F_SKY1")
				return false
			end

			if ((mo.z + FixedMul(16*FRACUNIT,mo.scale) < floorz)
				or (mo.z >= ceilingz))
				floorclimb = true
			end
		end

		if (not floorclimb)
			return false
		end

		return true
	else
		return true
	end

	return false
end

local function PTR_GlideClimbTraverse(slidemo,li)
	local checkline = li
	local rover
	local topheight, bottomheight
	local fofline = false
	local checksector = (li.backsector and not P_PointOnLineSide(slidemo.x, slidemo.y, li)) and li.backsector or li.frontsector

	if (checksector.ffloors)
		for rover in checksector.ffloors()
			if (not(rover.flags & FF_EXISTS) or not(rover.flags & FF_BLOCKPLAYER) or ((rover.flags & FF_BUSTUP) and (slidemo.player.charflags & SF_CANBUSTWALLS)))
				continue
			end

			topheight = rover.topheight
			bottomheight = rover.bottomheight

			if (topheight < slidemo.z)
				continue
			end

			if (bottomheight > slidemo.z + slidemo.height)
				continue
			end

			// Got this far, so I guess it's climbable. // TODO: Climbing check, also, better method to do this?
			if (rover.master.flags & ML_TFERLINE)
				local linenum = checksector.lines[0]
				checkline = rover.master.frontsector.lines[0] + linenum
				fofline = true
			end

			break
		end
	end

	// see about climbing on the wall
	if (not(checkline.flags & ML_NOCLIMB) and checkline.special ~= 41)
		local canclimb
		local climbangle, climbline
		local whichside = P_PointOnLineSide(slidemo.x, slidemo.y, li)

		climbangle = checkline.angle
		climbline = climbangle

		if (whichside) // on second side?
			climbline = $+ ANGLE_180
		end

		climbangle = $+ (ANGLE_90 * (whichside and -1 or 1))
		
		canclimb = (li.backsector and P_IsClimbingValid(slidemo.player, climbangle) or true)

		if canclimb
			return true,climbangle
		end
	end
	return false,0
end

addHook("MobjMoveBlocked", function(mo, mobj, line)
    if not (mo.player and mo.skin == "ovasonic"
    and not P_IsObjectOnGround(mo)) return end
	if mo.player.powers[pw_carry] == CR_NIGHTSMODE return end
	mo.player.lastlinehit = -1
	mo.player.lastsidehit = -1
    P_SlideMove(mo)
    mo.player.powers[pw_pushing] = 3
    local line = lines[mo.player.lastlinehit]
	if line and line.valid
		if line and line.valid
			if PTR_GlideClimbTraverse(mo,line)
			and mo.player and mo.player.valid
				local real,ang = PTR_GlideClimbTraverse(mo,line)
				if (P_IsClimbingValid(mo.player,ang) or (mobj and mobj.valid))
					if mobj and mobj.valid then ang = R_PointToAngle2(mo.x,mo.y,mobj.x,mobj.y) end
					mo.state = S_OVA_WALLRUN
					mo.player.pflags = $ & ~(PF_JUMPED|PF_THOKKED)
					mo.player.wallrun = true
					mo.player.wallangle = ang
					mo.player.drawangle = ang
					P_SetObjectMomZ(mo, mo.player.wallspeed, false)
				end
			end
		elseif mobj and mobj.valid
			local ang = R_PointToAngle2(mo.x,mo.y,mobj.x,mobj.y)
			mo.state = S_OVA_WALLRUN
			mo.player.pflags = $ & ~(PF_JUMPED|PF_THOKKED)
			mo.player.wallrun = true
			mo.player.wallangle = ang
			mo.player.drawangle = ang
			P_SetObjectMomZ(mo, mo.player.wallspeed, false)
		end
    end

    return true
end, MT_PLAYER)

local function canwalljump(p)
	return ((p.cmd.buttons&BT_JUMP and(p.pflags&~PF_JUMPED)) and p.jumptap and p.wallrun)
end

addHook("PlayerThink", function(p)
	if not(p.mo and p.mo.valid and p.mo.skin == "ovasonic") then return end
	
	if P_IsObjectOnGround(p.mo)
		p.wallrun = false
	end
	
	if p.cmd.buttons&BT_JUMP
		if not p.jumpdown
			p.jumptap = true
		else
			p.jumptap = false
		end
		p.jumpdown = true
	else
		p.jumptap = false
		p.jumpdown = false
	end
	
	if p.wallrun
		if p.wallspeed < 16*FRACUNIT
			p.wallspeed = $+FRACUNIT/8
		end
		if p.mo.state != S_OVA_WALLRUN
			p.wallrun = false
			p.wallspeed = FRACUNIT*4
		end
		if leveltime % 5 == 0
			local dust = P_SpawnMobjFromMobj(p.mo, 0,0,0, MT_SPINDUST)
			dust.momx = P_RandomRange(0,4)*p.mo.scale
			dust.momy = P_RandomRange(0,4)*p.mo.scale
			dust.momz = -p.mo.momz/2
			dust.flags = $|MF_NOCLIPHEIGHT
			S_StartSound(p.mo, sfx_s3k7e)
		end
		if canwalljump(p)
			p.mo.state = S_PLAY_JUMP
			S_StartSound(p.mo, sfx_jump)
			p.mo.momz = FRACUNIT*12
			P_InstaThrust(p.mo, p.wallangle+ANGLE_180, FRACUNIT*15)
			p.pflags = $|PF_JUMPED
			p.wallrun = false
		end
	else
		p.wallspeed = p.speed/4
	end
end)